#include <algorithm>
#include <vector>
#include <array>
#include <chrono>
#include <cstdio>
#include <cassert>

const size_t size = 1<<25;

std::chrono::high_resolution_clock::time_point timer_start;
typedef std::vector<int>::iterator iterator;

void start() {
    timer_start = std::chrono::high_resolution_clock::now();
}

void stop(const char* message) {
    float elapsedTime = std::chrono::duration_cast<std::chrono::duration<float>>(std::chrono::high_resolution_clock::now() - timer_start).count();
    printf("%s %f\n", message, elapsedTime);
    start();
}

void check_sorted(int* begin, int* end) {
    for (int *p=begin+1; p < end; p++) {
        assert(*(p-1) <= *p);
    }
}

template<typename T>
inline void bubblesort(T begin, T end) {
    while (end > begin) {
        end -= 1;
        for (T p=begin; p<end; p++) {
            if (*p > *(p+1)) {
                std::swap(*p, *(p+1));
            }
        }
    }
}

inline int guess_median(iterator begin, iterator end) {
    iterator a = begin;
    iterator c = begin + (end - begin) / 2;
    iterator b = end - 1;
    if (*c < *a) std::swap(a, c);
    if (*c < *b) std::swap(*c, *b);
    else if (*a > *b) std::swap(*a, *b);
    assert(*a <= *b);
    assert(*b <= *c);
    return *b;
}

inline iterator partition_onepass(iterator begin, iterator end, int pivot) {
    for (iterator mid = begin; mid < end; mid++) {
        if (*mid < pivot) {
            std::swap(*mid, *begin);
            begin += 1;
        }
    }
    return begin;
}

inline iterator partition_twosided(iterator begin, iterator end, int pivot) {
    iterator left = begin, right = end - 1;
    while (true) {
        while (*left < pivot) {
            left += 1;
        }
        while (*right >= pivot && right > left) {
            right -= 1;
        }
        if (right <= left) {
            return left;
        }
        std::swap(*left, *right);
        left += 1;
        right -= 1;
    }
}

// place a range [end - n..end-1] to [begin..begin+n]
inline iterator swap_pivots(iterator begin, iterator end) {
// TODO!
}
const size_t bubblesort_limit = 20;

size_t count_calls = 0;
void quicksort(iterator begin, iterator end) {
    count_calls += 1;
    if (end - begin <= 1) {
        return;
    } else if (end - begin <= bubblesort_limit) {
        bubblesort(begin, end);
        return;
    }
    std::swap(begin[(end - begin)/2], end[-1]);
    int pivot = end[-1];
    iterator mid_end = end - 1;
    iterator mid_start;
    if (0) {
        mid_start = partition_twosided(begin, mid_end, pivot);
    } else if (1) {
        mid_start = partition_onepass(begin, mid_end, pivot);
    } else {
        mid_start = std::partition(begin, end, [pivot](int i){ return i < pivot; });
    }
    quicksort(begin, mid_start);
    for (; mid_end < end; mid_start++, mid_end++) {
        std::swap(*mid_end, *mid_start);
    }
    quicksort(mid_end, end);
}

struct Job {
    iterator begin, end;
    Job(iterator _begin, iterator _end): begin(_begin), end(_end) {};
};

void stacksort(std::vector<Job> stack) {
    while (!stack.empty()) {
        count_calls += 1;
        iterator begin = stack.back().begin;
        iterator end = stack.back().end;
        stack.pop_back();
        if (end - begin > bubblesort_limit) {
            int pivot = guess_median(begin, end);
            iterator mid = partition_onepass(begin, end, pivot);
            if (mid - begin > 1) {
                stack.emplace_back(begin, mid);
            }
            if (end - mid > 2) {
                stack.emplace_back(mid + 1, end);
            }
        } else {
            bubblesort(begin, end);
        }
    }
}

void stacksort(iterator begin, iterator end) {
    std::vector<Job> stack;
    stack.emplace_back(begin, end);
    stacksort(stack);
}

int main() {
    std::vector<int> data;
    for (size_t i=0; i<size; i++) {
        data.push_back((i%7)*(i%1234)*(i%4321));
    }
    start();
    if (0) {
        std::sort(data.begin(), data.end());
        stop("std::sort");
    } else if (1) {
        quicksort(data.begin(), data.end());
        stop("recursive quicksort");
    } else {
        stacksort(data.begin(), data.end());
        stop("stack quicksort");
    }
    printf("%lu calls on %lu elements\n", count_calls, size);
    if (size < 100) {
        for (size_t i=0; i<size; i++) {
            printf("%d ", data[i]);
        }
        printf("\n");
    } else {
        check_sorted(&data[0], &data[size - 1]);
    }
}